class Base():
    myfoobar = 3


class X(Base):
    def func(self, foo):
        pass


class Y(X):
    def actual_function(self):
        pass

    #? []
    def actual_function
    #? ['func']
    def f

    #? []
    def __class__

    #? ['__repr__']
    def __repr__

    #? []
    def mro

    #? ['myfoobar']
    myfoobar

#? []
myfoobar

# -----------------
# Inheritance
# -----------------

class Super():
    enabled = True
    if enabled:
        yo_dude = 4

class Sub(Super):
    #? ['yo_dude']
    yo_dud
